# Andrey version 1.0.0
# AndreyChecker.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Andrey, a pedagogical implementation of Markov chains.
# Andrey is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Andrey is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Andrey.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Undefined references
# 4) Specific checks

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLChecker import *
from AndreyModels import *

# 2) Creator
# ----------

class AndreyChecker(S2MLChecker):
  def __init__(self, engine):
    S2MLChecker.__init__(self, engine)

# 3) Undefined references
# -----------------------

  def CheckModel(self, model):
    S2MLChecker.CheckModel(self, model)
    if self._errors:
      return self._errors
    self.CheckEventsInBlock(model)
    self.ResetTimeTypesOfStatesInBlock(model)
    self.CheckTimeTypesOfStatesInBlock(model)
    self.CompleteTypesOfStatesInBlock(model)
    return self._errors

  def CheckReferencesInBlock(self, block):
    if block.GetType()==AndreyType.BLOCK and block.GetSubType()==AndreyType.UNKNOWN:
      block.SetSubType(AndreyType.MARKOV_SYSTEM)
    for element in block.GetElements():
      if AndreyModel.IsBlock(element):
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckReferencesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif AndreyModel.IsMarkovChainState(element):
        self._checkedElement = element
        self._CheckReferencesInAttributesOfElement(element)
        self._checkedElement = None
      elif AndreyModel.IsMarkovChainEvent(element):
        self._checkedElement = element
        self._CheckReferencesInAttributesOfElement(element)
        self._checkedElement = None
      elif AndreyModel.IsParameter(element):
        self._checkedElement = element
        self._CheckReferencesInStochasticExpression(element.GetDefinition())
        self._checkedElement = None
    return self._errors

# 4) Specific checks
# ------------------

  def CheckEventsInBlock(self, block):
    for element in block.GetElements():
      if AndreyModel.IsBlock(element):
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckEventsInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif AndreyModel.IsMarkovChainEvent(element):
        self._checkedElement = element
        self.CheckMarkovChainEvent(element)
        self._checkedElement = None
    return self._errors

  def CheckMarkovChainEvent(self, event):
    expression = event.GetAttribute("rate")
    if expression==None:
      expression = event.GetAttribute("probability")
      if expression!=None:
        event.SetTimeType(AndreyType.IMMEDIATE)
    else:
      event.SetTimeType(AndreyType.TEMPORIZED)
    if expression==None:
      self._RaiseError("In block %s, no rate or probability associated with event %s" % (self._checkedBlock.GetAbsolutePath().ToString(), self._checkedElement.GetPath().ToString()))
      return

  def ResetTimeTypesOfStatesInBlock(self, block):
    for element in block.GetElements():
      if AndreyModel.IsBlock(element):
        self.ResetTimeTypesOfStatesInBlock(element)
      elif AndreyModel.IsMarkovChainState(element):
        element.SetTimeType(AndreyType.UNKNOWN)

  def CheckTimeTypesOfStatesInBlock(self, block):
    if AndreyModel.IsMarkovChainBlock(block):
      block.SetTimeType(AndreyType.IMMEDIATE)
    for element in block.GetElements():
      if AndreyModel.IsBlock(element):
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckTimeTypesOfStatesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif AndreyModel.IsMarkovChainTransition(element):
        event = element.GetEvent()
        sourceState = element.GetSourceState()
        if event.GetTimeType()==AndreyType.TEMPORIZED:
          block.SetTimeType(AndreyType.TEMPORIZED)
          if sourceState.GetTimeType()==AndreyType.UNKNOWN:
            sourceState.SetTimeType(AndreyType.TEMPORIZED)
          elif sourceState.GetTimeType()==AndreyType.IMMEDIATE:
            self._RaiseError("In block %s, type conflict in out going transitions of event %s." % (self._checkedBlock.GetAbsolutePath().ToString(), sourceState.GetPath().ToString()))
        else:
          if sourceState.GetTimeType()==AndreyType.UNKNOWN:
            sourceState.SetTimeType(AndreyType.IMMEDIATE)
          elif sourceState.GetTimeType()==AndreyType.TEMPORIZED:
            self._RaiseError("In block %s, type conflict in out going transitions of event %s." % (self._checkedBlock.GetAbsolutePath().ToString(), sourceState.GetPath().ToString()))
    return self._errors

  def CompleteTypesOfStatesInBlock(self, block):
    for element in block.GetElements():
      if AndreyModel.IsBlock(element):
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CompleteTypesOfStatesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif AndreyModel.IsMarkovChainState(element):
        if element.GetTimeType()==AndreyType.UNKNOWN:
          element.SetTimeType(block.GetTimeType())
