# Andrey version 1.0.0
# AndreyEngine.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Andrey, a pedagogical implementation of Markov chains.
# Andrey is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Andrey is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Andrey.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Settings
# 4) Algorithms
# 5) Methods to assess discrete-time Markov chains
# 6) Methods to assess continuous-time Markov chains
# 7) Initialization of chains
# 8) Matrix exponentiation
# 9) Management of result matrix
# 10) Assessment of arithmetic expressions

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLEngine import *
from AndreyModels import *
from AndreyLexer import *
from AndreyParser import *
from AndreyPrinter import *
from AndreyFlattener import *
from AndreyChecker import *
from AndreyInterpreter import *

# 2) Creator
# ----------

class AndreyEngine(S2MLEngine):
  def __init__(self):
    S2MLEngine.__init__(self)
    self._interpreter = None
    self._sourceModel = AndreyModel()
    self._sourceNode = None
    self._targetNode = None
    self._matrix = None
    self._chains = []
    self._transitions = []
    self._states = []
    self._rewards = []
    self._numbersOfIterations = []
    self._currentNumberOfIterations = -1
    self._missionTimes = []
    self._currentMissionTime = S2MLReal.New(-1.0)
    self._probabilities = []
    self._sojournTimes = []
    self._P0 = []
    self._P1 = []
    self._K0 = []
    self._K1 = []
    self._Q = []
    self._assessPointReward = True
    self._matrix = None
    self._probabilityIndex = 0
    self._sojournIndex = 0
    self._pointRewardIndex = 0
    self._meanRewardIndex = 0
    self._dtRatio = S2MLReal.New(1.0)
    self._epsilon = S2MLReal.New(1.0e-20)
    self._startTime = S2MLReal.New(0.0)
    self._endTime = S2MLReal.New(0.0)

# 3) Settings
# -----------

  def CreateTools(self):
    self._lexer = AndreyLexer(self)
    self._parser = AndreyParser(self)
    self._expressionManager = S2MLStochasticExpressionManager(self)
    self._expressionManager.RegisterOperator(AndreyType.IN_EXPRESSION, AndreyType.KEYWORD_IN, 2, 2, AndreyEngine.CalculateValueOfInExpression)
    self._printer = AndreyPrinter(self)
    self._parser = AndreyParser(self)
    self._flattener = AndreyFlattener(self)
    self._checker = AndreyChecker(self)
    self._interpreter = AndreyInterpreter(self)

  def GetDtRatio(self):
    return self._dtRatio
  def SetDtRatio(self, dtRatio):
    self._dtRatio = dtRatio

# 4) Methods to import, instantiate and export models
# ---------------------------------------------------

  def ResetModel(self):
    self._sourceModel = AndreyModel()
    self._targetModel = None
    self._targetBlock = None
    return 0

# 5) Methods to assess discrete-time Markov chains
# ------------------------------------------------

  def AssessDiscreteTimeMarkovSystem(self, blockName, numbersOfIterations, printProbabilities, printPointRewards, fileName, mode):
    if self._targetModel==None:
      self.RaiseError("No instantiated model")
      return 1
    self._targetBlock = self._targetModel.LookForBlock(blockName)
    if self._targetBlock==None:
      self.RaiseError("No block %s in instantiated model" % blockName)
      return 1
    if not self._targetBlock.IsDiscrete():
      self.RaiseError("Block %s is not a discrete model" % blockName)
      return 1

    # Creation of the mission
    self._CreateDiscreteTimeMission(numbersOfIterations)

    # Collection of states, transitions and rewards.
    self._states = []
    self._transitions = []
    self._rewards = []
    self._CollectMarkovChainElements(self._targetBlock)
    # print("block\t" + self._targetBlock.GetPath().ToString())
    # print("#states\t" + str(len(self._states)))
    # print("#transitions\t" + str(len(self._transitions)))
    # print("#rewards\t" + str(len(self._rewards)))

    # Creation of vectors of probabilities and sojourn times
    self._probabilities = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    self._Q = [S2MLReal.New(0.0) for i in range(0, len(self._states))]

    # Initialization
    self._expressionManager.ResetValuesOfParameters(self._targetBlock)
    self._CalculateRatesOfTemporizedTransitionsAndProbabilitiesOfImmediateTransitions()
    self._CalculateInitialProbalilitiesOfStates()

    # Creation of result matrix
    self._CreateResultMatrix(True, printProbabilities, False, printPointRewards, False)

    # Main loop
    previousNumberOfIterations = 0
    for numberOfIterationsIndex in range(0, len(self._numbersOfIterations)):
      currentNumberOfIterations = self._numbersOfIterations[numberOfIterationsIndex]
      self._AssessDiscreteTimeMarkovChain(self._targetBlock, previousNumberOfIterations, currentNumberOfIterations)
      self._UpdateMatrix(True, numberOfIterationsIndex, printProbabilities, False, printPointRewards, False)
      previousNumberOfIterations = currentNumberOfIterations

    # Printing of the results
    self._printer.ExportMatrix(self._matrix, fileName, mode)
    return 0

  def _CreateDiscreteTimeMission(self, numbersOfIterations):
    self._numbersOfIterations = numbersOfIterations.copy()
    self._numbersOfIterations.sort()
    previousNumberOfIterations = -1
    index = 0
    while index<len(self._numbersOfIterations):
      currentNumberOfIterations = self._numbersOfIterations[index]
      if currentNumberOfIterations<0 or currentNumberOfIterations==previousNumberOfIterations:
        self._numbersOfIterations.pop(index)
      else:
        previousNumberOfIterations = currentNumberOfIterations
        index += 1

  def _AssessDiscreteTimeMarkovChain(self, chain, startNumberOfIterations, endNumberOfIterations):
    convergence = False
    for numberOfIterations in range(startNumberOfIterations, endNumberOfIterations):
      self._FillVectorWithVector(self._Q, self._probabilities, chain.GetStartIndexStates(), chain.GetEndIndexStates())
      self._MultiplyRightProbabilityMatrixByColumnVector(self._probabilities, self._Q, chain.GetStartIndexStates(), chain.GetEndIndexStates(), chain.GetStartIndexImmediateTransitions(), chain.GetEndIndexTransitions())
      if self._AreVectorsAlmostEqual(self._probabilities, self._Q, chain.GetStartIndexStates(), chain.GetEndIndexStates()):
        convergence = True
        break

# 6) Methods to assess continuous-time Markov chains
# --------------------------------------------------

  def AssessContinuousTimeMarkovSystem(self, blockName, missionTimes, printProbabilities, printSojournTimes, printPointRewards, printMeanRewards, fileName, mode):
    if self._targetModel==None:
      self.RaiseError("No instantiated model")
      return 1
    self._targetBlock = self._targetModel.LookForBlock(blockName)
    if self._targetBlock==None:
      self.RaiseError("No block %s in instantiated model" % blockName)
      return 1
    if not self._targetBlock.IsContinuous():
      self.RaiseError("Block %s is not a continuous model" % blockName)
      return 1

    # Creation of the mission
    self._CreateContinuousTimeMission(missionTimes)

    # Collection of states, transitions and rewards.
    self._states = []
    self._transitions = []
    self._rewards = []
    self._CollectMarkovChainElements(self._targetBlock)
    # print("block\t" + self._targetBlock.GetPath().ToString())
    # print("#states\t" + str(len(self._states)) + " " + str(self._targetBlock.GetStartIndexStates()) + " " + str(self._targetBlock.GetStartIndexImmediateStates()) + " " + str(self._targetBlock.GetEndIndexStates()))
    # print("#transitions\t" + str(len(self._transitions)))
    # print("#rewards\t" + str(len(self._rewards)))

    # Creation of vectors of probabilities and sojourn times
    self._probabilities = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    self._sojournTimes = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    self._P0 = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    self._P1 = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    self._K0 = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    self._K1 = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    self._Q = [S2MLReal.New(0.0) for i in range(0, len(self._states))]

    # Initialization
    self._expressionManager.ResetValuesOfParameters(self._targetBlock)
    self._CalculateRatesOfTemporizedTransitionsAndProbabilitiesOfImmediateTransitions()
    self._CalculateInitialProbalilitiesOfStates()
    self._CalculateMaximumRate(self._targetBlock)

    # Creation of result matrix
    self._CreateResultMatrix(False, printProbabilities, printSojournTimes, printPointRewards, printMeanRewards)

    # Main loop
    previousMissionTime = S2MLReal.New(0.0)
    for missionTimeIndex in range(0, len(self._missionTimes)):
      currentMissionTime = self._missionTimes[missionTimeIndex]
      self._AssessContinuousTimeMarkovChainByMatrixExponentiation(self._targetBlock, previousMissionTime, currentMissionTime)
      self._UpdateMatrix(False, missionTimeIndex, printProbabilities, printSojournTimes, printPointRewards, printMeanRewards)
      previousMissionTime = currentMissionTime

    # Printing of the results
    self._printer.ExportMatrix(self._matrix, fileName, mode)
    return 0

  def _CreateContinuousTimeMission(self, missionTimes):
    self._missionTimes = missionTimes.copy()
    self._missionTimes.sort()
    previousMissionTime = -1
    index = 0
    while index<len(self._missionTimes):
      currentMissionTime = self._missionTimes[index]
      if currentMissionTime<0 or currentMissionTime==previousMissionTime:
        self._missionTimes.pop(index)
      else:
        previousMissionTime = currentMissionTime
        index += 1

# 7) Initialization of chains
# ---------------------------

  def _CollectMarkovChainElements(self, block):
    block.SetStartIndexStates(len(self._states))
    block.SetStartIndexTransitions(len(self._transitions))
    for element in block.GetElements():
      if AndreyModel.IsMarkovChainState(element) and element.GetTimeType()==AndreyType.TEMPORIZED:
        element.SetIndex(len(self._states))
        self._states.append(element)
      elif AndreyModel.IsMarkovChainTransition(element) and element.GetTimeType()==AndreyType.TEMPORIZED:
        element.SetIndex(len(self._transitions))
        self._transitions.append(element)
      elif AndreyModel.IsMarkovChainReward(element):
        element.SetIndex(len(self._rewards))
        self._rewards.append(element)
    block.SetStartIndexImmediateStates(len(self._states))
    block.SetStartIndexImmediateTransitions(len(self._transitions))
    for element in block.GetElements():
      if AndreyModel.IsMarkovChainState(element) and element.GetTimeType()==AndreyType.IMMEDIATE:
        element.SetIndex(len(self._states))
        self._states.append(element)
      elif AndreyModel.IsMarkovChainTransition(element) and element.GetTimeType()==AndreyType.IMMEDIATE:
        element.SetIndex(len(self._transitions))
        self._transitions.append(element)
    block.SetEndIndexStates(len(self._states))
    block.SetEndIndexTransitions(len(self._transitions))

  def _CalculateRatesOfTemporizedTransitionsAndProbabilitiesOfImmediateTransitions(self):
    for index in range(0, len(self._transitions)):
      transition = self._transitions[index]
      event = transition.GetEvent()
      if event.GetTimeType()==AndreyType.TEMPORIZED:
        event.SetRate(S2MLReal.New(-1.0))
        event.SetProbability(S2MLReal.New(0.0))
      else:
        event.SetRate(S2MLReal.New(0.0))
        event.SetProbability(S2MLReal.New(-1.0))
    for index in range(0, len(self._transitions)):
      transition = self._transitions[index]
      event = transition.GetEvent()
      if event.GetTimeType()==AndreyType.TEMPORIZED:
        rate = event.GetRate()
        if rate==S2MLReal.New(-1.0):
          expression = event.GetAttribute("rate")
          value = self._expressionManager.CalculateValueOfExpression(expression)
          rate = value.ToFloat()
          event.SetRate(rate)
        transition.SetRate(rate)
        transition.SetProbability(S2MLReal.New(0.0))
        # print("rate(" + event.GetIdentifier() + ") = " + str(rate))
      else:
        probability = event.GetProbability()
        if probability==S2MLReal.New(-1.0):
          expression = event.GetAttribute("probability")
          value = self._expressionManager.CalculateValueOfExpression(expression)
          probability = value.ToFloat()
          event.SetProbability(probability)
          # print("p(" + event.GetIdentifier() + ") = " + str(probability))
        transition.SetRate(S2MLReal.New(0.0))
        transition.SetProbability(probability)

  def _CalculateInitialProbalilitiesOfStates(self):
    for index in range(0, len(self._states)):
      state = self._states[index]
      expression = state.GetAttribute("probability")
      if expression==None:
        probability = S2MLReal.New(0.0)
      else:
        value = self._expressionManager.CalculateValueOfExpression(expression)
        probability = value.ToFloat()
        # print("p(" + state.GetIdentifier() + ") = " + str(probability))
      state.SetProbability(probability)
      self._probabilities[index] = probability

  def _CalculateMaximumRate(self, chain):
    diagonal = [S2MLReal.New(0.0) for i in range(0, len(self._states))]
    for index in range(chain.GetStartIndexTransitions(), chain.GetStartIndexImmediateTransitions()):
      transition = self._transitions[index]
      sourceState = transition.GetSourceState()
      diagonal[sourceState.GetIndex()] += transition.GetRate()
    maxRate = S2MLReal.New(0.0)
    for index in range(chain.GetStartIndexStates(), chain.GetStartIndexImmediateStates()):
      if diagonal[index]>maxRate:
        maxRate = diagonal[index]
    chain.SetMaximumRate(maxRate)

# 8) Matrix exponentiation
# ------------------------

  def _AssessContinuousTimeMarkovChainByMatrixExponentiation(self, chain, startTime, endTime):
    dt = self._SelectDt(chain, startTime, endTime)
    self._CalculateProbabilitiesOfTemporizedTransitions(chain, dt)
    self._FillVectorWithVector(self._P0, self._probabilities, chain.GetStartIndexStates(), chain.GetEndIndexStates())
    self._FillVectorWithVector(self._K0, self._probabilities, chain.GetStartIndexStates(), chain.GetEndIndexStates())
    convergence = False
    time = startTime
    while True:
      if time>=endTime:
        break
      time = S2MLReal.Add(time, dt)
      n = 0
      while True:
        n += 1
        self._FillVectorWithValue(self._K1, 0.0, chain.GetStartIndexStates(), chain.GetEndIndexStates())
        # self._FillVectorWithVector(self._K1, self._K0, chain.GetStartIndexStates(), chain.GetEndIndexStates());
        self._MultiplyRightProbabilityMatrixByColumnVector(self._K1, self._K0, chain.GetStartIndexStates(), chain.GetEndIndexStates(), chain.GetStartIndexTransitions(), chain.GetStartIndexImmediateTransitions())
        self._DivideVectorByValue(self._K1, n, chain.GetStartIndexStates(), chain.GetEndIndexStates())
        self._FillVectorWithSumOfVectors(self._P1, self._P0, self._K1, chain.GetStartIndexStates(), chain.GetEndIndexStates())
        if self._AreVectorsAlmostEqual(self._P0, self._P1, chain.GetStartIndexStates(), chain.GetEndIndexStates()):
          convergence = True
          break
        self._FillVectorWithVector(self._P0, self._P1, chain.GetStartIndexStates(), chain.GetEndIndexStates())
        self._FillVectorWithVector(self._K0, self._K1, chain.GetStartIndexStates(), chain.GetEndIndexStates());
      if chain.HasImmediateState():
        self._FlushImmediateStates(self._P1, self._Q, chain)
      self._AddMeanOfVectorsToVector(self._sojournTimes, self._probabilities, self._P1, dt, chain.GetStartIndexStates(), chain.GetStartIndexImmediateStates());
      convergence = self._AreVectorsAlmostEqual(self._probabilities, self._P1, chain.GetStartIndexStates(), chain.GetEndIndexStates())
      self._FillVectorWithVector(self._probabilities, self._P1, chain.GetStartIndexStates(), chain.GetEndIndexStates());
      if convergence:
        break
      self._FillVectorWithVector(self._P0, self._probabilities, chain.GetStartIndexStates(), chain.GetEndIndexStates());
      self._FillVectorWithVector(self._K0, self._probabilities, chain.GetStartIndexStates(), chain.GetEndIndexStates());
    if convergence:
      remainingTime = endTime - time;
      self._AddVectorToVector(self._sojournTimes, self._probabilities, remainingTime, chain.GetStartIndexStates(), chain.GetStartIndexImmediateStates());

  def _SelectDt(self, chain, startTime, endTime):
    missionTime = S2MLReal.Sub(endTime, startTime)
    if chain.GetMaximumRate()==S2MLReal.New(0.0):
      return missionTime 
    dt = S2MLReal.Div(self._dtRatio, chain.GetMaximumRate())
    if dt>=missionTime:
      return missionTime 
    n = S2MLReal.Ceil(S2MLReal.Div(missionTime, dt));
    return S2MLReal.Div(missionTime, n);

  def _CalculateProbabilitiesOfTemporizedTransitions(self, chain, dt):
    for index in range(chain.GetStartIndexTransitions(), chain.GetStartIndexImmediateTransitions()):
      transition = self._transitions[index]
      rate = transition.GetRate()
      probability = S2MLReal.Mul(rate, dt)
      transition.SetProbability(probability)

  def _FillVectorWithValue(self, V, value, startIndex, endIndex):
    for index in range(startIndex, endIndex):
      V[index] = value

  def _DivideVectorByValue(self, V, value, startIndex, endIndex):
    for index in range(startIndex, endIndex):
      V[index] = S2MLReal.Div(V[index], value)

  def _FillVectorWithVector(self, V, W, startIndex, endIndex):
    for index in range(startIndex, endIndex):
      V[index] = W[index]

  def _FillVectorWithSumOfVectors(self, V, W1, W2, startIndex, endIndex):
    for index in range(startIndex, endIndex):
      V[index] = S2MLReal.Add(W1[index], W2[index])

  def _AddVectorToVector(self, V, W, coefficient, startIndex, endIndex):
    for index in range(startIndex, endIndex):
      V[index] = S2MLReal.Add(V[index], S2MLReal.Mul(coefficient, W[index]))

  def _AddMeanOfVectorsToVector(self, V, W1, W2, coefficient, startIndex, endIndex):
    for index in range(startIndex, endIndex):
      V[index] = S2MLReal.Add(V[index], S2MLReal.Div(S2MLReal.Mul(coefficient, S2MLReal.Add(W1[index], W2[index])), S2MLReal.New(2.0)))

  def _AreVectorsAlmostEqual(self, V, W, startIndex, endIndex):
    for index in range(startIndex, endIndex):
      if not S2MLReal.AlmostEq(V[index], W[index], self._epsilon):
        return False
    return True

  def _MultiplyRightProbabilityMatrixByColumnVector(self, V, W, startIndexStates, endIndexStates, startIndexTransitions, endIndexTransitions):
    # self._FillVectorWithValue(V, 0.0, startIndexStates, endIndexStates);
    # self._FillVectorWithVector(W, V, startIndexStates, endIndexStates);
    for transitionIndex in range(startIndexTransitions, endIndexTransitions):
      transition = self._transitions[transitionIndex]
      sourceState = transition.GetSourceState()
      sourceStateIndex = sourceState.GetIndex()
      targetState = transition.GetTargetState()
      targetStateIndex = targetState.GetIndex()
      if sourceState!=targetState:
        transferredProbability = S2MLReal.Mul(transition.GetProbability(), W[sourceStateIndex])
        V[sourceStateIndex] = S2MLReal.Sub(V[sourceStateIndex], transferredProbability)
        V[targetStateIndex] = S2MLReal.Add(V[targetStateIndex], transferredProbability)

  def _FlushImmediateStates(self, V, W, chain):
    while True:
      numberOfActiveImmediateStates = 0
      for stateIndex in range(chain.GetStartIndexImmediateStates(), chain.GetEndIndexStates()):
        W[stateIndex] = V[stateIndex]
        if V[stateIndex]>0.0:
          numberOfActiveImmediateStates += 1
      if numberOfActiveImmediateStates==0:
        break
      for transitionIndex in range(chain.GetStartIndexImmediateTransitions(), chain.GetEndIndexTransitions()):
        transition = self._transitions[transitionIndex]
        sourceState = transition.GetSourceState()
        sourceStateIndex = sourceState.GetIndex()
        if W[sourceStateIndex]>0.0:
          targetState = transition.GetTargetState()
          targetStateIndex = targetState.GetIndex()
          transferredProbability = S2MLReal.Mul(transition.GetProbability(), W[sourceStateIndex])
          V[sourceStateIndex] = S2MLReal.Sub(V[sourceStateIndex], transferredProbability)
          V[targetStateIndex] = S2MLReal.Add(V[targetStateIndex], transferredProbability)

# 9) Management of result matrix
# ------------------------------

  def _CreateResultMatrix(self, isDiscreteTimeMission, printProbabilities, printSojournTimes, printPointRewards, printMeanRewards):
    self._matrix = S2MLMatrix()
    row = self._matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("Model"))
    row.Put(1, S2MLConstant.NewString(self._targetBlock.GetIdentifier()))
    self._firstFreeMatrixIndex = 0
    if printProbabilities:
      row = self._matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString("Probabilities"))
      self._CreateRowOfMissionSteps(row, isDiscreteTimeMission)
      self._probabilityIndex = self._matrix.GetNumberOfRows()
      for state in self._states:
        row = self._matrix.AppendRow()
        absolutePath = state.GetAbsolutePath()
        absolutePath.RemoveFirstIdentifier()
        row.Put(0, S2MLConstant.NewString(absolutePath.ToString()))
        self._CreateRowForValuesAtMissionSteps(row, isDiscreteTimeMission)
    if printSojournTimes:
      row = self._matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString("Sojourn-times"))
      self._CreateRowOfMissionSteps(row, isDiscreteTimeMission)
      self._sojournTimeIndex = self._matrix.GetNumberOfRows()
      for state in self._states:
        row = self._matrix.AppendRow()
        absolutePath = state.GetAbsolutePath()
        absolutePath.RemoveFirstIdentifier()
        row.Put(0, S2MLConstant.NewString(absolutePath.ToString()))
        self._CreateRowForValuesAtMissionSteps(row, isDiscreteTimeMission)
    if printPointRewards:
      row = self._matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString("Point-rewards"))
      self._CreateRowOfMissionSteps(row, isDiscreteTimeMission)
      self._pointRewardIndex = self._matrix.GetNumberOfRows()
      for reward in self._rewards:
        row = self._matrix.AppendRow()
        absolutePath = reward.GetAbsolutePath()
        absolutePath.RemoveFirstIdentifier()
        row.Put(0, S2MLConstant.NewString(absolutePath.ToString()))
        self._CreateRowForValuesAtMissionSteps(row, isDiscreteTimeMission)
    if printMeanRewards:
      row = self._matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString("Mean-rewards"))
      self._CreateRowOfMissionSteps(row, isDiscreteTimeMission)
      self._meanRewardIndex = self._matrix.GetNumberOfRows()
      for reward in self._rewards:
        row = self._matrix.AppendRow()
        absolutePath = reward.GetAbsolutePath()
        absolutePath.RemoveFirstIdentifier()
        row.Put(0, S2MLConstant.NewString(absolutePath.ToString()))
        self._CreateRowForValuesAtMissionSteps(row, isDiscreteTimeMission)

  def _CreateRowOfMissionSteps(self, row, isDiscreteTimeMission):
    if isDiscreteTimeMission:
      for i in range(0, len(self._numbersOfIterations)):
        row.Put(i+1, S2MLConstant.NewInteger(self._numbersOfIterations[i]))
    else:
      for i in range(0, len(self._missionTimes)):
        row.Put(i+1, S2MLConstant.NewReal(self._missionTimes[i]))

  def _CreateRowForValuesAtMissionSteps(self, row, isDiscreteTimeMission):
    if isDiscreteTimeMission:
      for i in range(0, len(self._numbersOfIterations)):
        row.Put(i+1, S2MLConstant.NewReal(0.0))
    else:
      for i in range(0, len(self._missionTimes)):
        row.Put(i+1, S2MLConstant.NewReal(0.0))

  def _UpdateMatrix(self, isChainDiscrete, stepIndex, printProbabilities, printSojournTimes, printPointRewards, printMeanRewards):
    if printProbabilities:
      for state in self._states:
        index = state.GetIndex()
        self._matrix.Set(self._probabilityIndex+index, stepIndex+1, S2MLConstant.NewReal(self._probabilities[index]))
    if printSojournTimes:
      for state in self._states:
        index = state.GetIndex()
        self._matrix.Set(self._sojournTimeIndex+index, stepIndex+1, S2MLConstant.NewReal(self._sojournTimes[index]))
    if printPointRewards:
      self. _assessPointReward = True
      for reward in self._rewards:
        reward.ResetValue()
      for reward in self._rewards:
        index = reward.GetIndex()
        constant = self._expressionManager.CalculateValueOfParameter(reward)
        value = constant.ToFloat()
        self._matrix.Set(self._pointRewardIndex+index, stepIndex+1, S2MLConstant.NewReal(value))
    if printMeanRewards:
      self._assessPointReward = False
      for reward in self._rewards:
        reward.ResetValue()
      self._currentMissionTime = self._missionTimes[stepIndex]
      for reward in self._rewards:
        index = reward.GetIndex()
        if self._currentMissionTime==0.0:
          value = 0.0
        else:
          constant = self._expressionManager.CalculateValueOfParameter(reward)
          value = constant.ToFloat()
        self._matrix.Set(self._meanRewardIndex+index, stepIndex+1, S2MLConstant.NewReal(value))

# 10) Assessment of arithmetic expressions
# ----------------------------------------

  @staticmethod
  def CalculateValueOfInExpression(expressionManager, expression):
    if expression.GetType()!=AndreyType.IN_EXPRESSION:
      return S2MLConstant.NewReal(0.0)
    engine = expressionManager.GetEngine()
    child0 = expression.GetChild(0)
    state = child0.GetReferredElement()
    index = state.GetIndex()
    if engine._assessPointReward:
      stateValue = engine._probabilities[index];
    else:
      stateValue = S2MLReal.Div(engine._sojournTimes[index], engine._currentMissionTime);
    child1 = expression.GetChild(1)
    childValue1 = expressionManager.CalculateValueOfExpression(child1)
    value = expression.GetValue()
    value.SetAsReal(stateValue)
    value.MulMe(childValue1)
    return value
