# Andrey version 1.0.0
# AndreyPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Andrey, a pedagogical implementation of Markov chains.
# Andrey is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Andrey is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Andrey.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
#   3.1) Models
#   3.2) Blocks and classes
#   3.3) Specific elements

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLPrinter import *
from AndreyModels import *

# 2) Creator
# ----------

class AndreyPrinter(S2MLPrinter):
  def __init__(self, engine):
    S2MLPrinter.__init__(self, engine)
    self._inTransition = False

# 3) Models
# ---------

# 3.1) Models
# -----------

  def PrintElement(self, element, output):
    if element.GetType()==AndreyType.BLOCK:
      self._UpdateLocation(False, output)
      self.PrintBlock(element, output)
    elif element.GetType()==AndreyType.CLASS:
      self._UpdateLocation(False, output)
      self.PrintClass(element, output)
    elif element.GetType()==AndreyType.INSTANCE:
      self._UpdateLocation(False, output)
      self.PrintInstance(element, output)
    elif element.GetType()==AndreyType.CLONES:
      self._UpdateLocation(False, output)
      self.PrintClonesDirective(element, output)
    elif element.GetType()==AndreyType.EXTENDS:
      self._UpdateLocation(False, output)
      self.PrintExtendsDirective(element, output)
    elif element.GetType()==AndreyType.EMBEDS:
      self._UpdateLocation(False, output)
      self.PrintEmbedsDirective(element, output)
    elif element.GetType()==AndreyType.MARKOV_CHAIN_STATE:
      self._UpdateLocation(False, output)
      self.PrintMarkovChainState(element, output)
    elif element.GetType()==AndreyType.MARKOV_CHAIN_EVENT:
      self._UpdateLocation(False, output)
      self.PrintMarkovChainEvent(element, output)
    elif element.GetType()==AndreyType.PARAMETER and element.GetSubType()==AndreyType.MARKOV_CHAIN_PARAMETER:
      self.PrintMarkovChainParameter(element, output)
    elif element.GetType()==AndreyType.PARAMETER and element.GetSubType()==AndreyType.MARKOV_CHAIN_REWARD:
      self._UpdateLocation(False, output)
      self.PrintMarkovChainReward(element, output)
    elif element.GetType()==AndreyType.MARKOV_CHAIN_TRANSITION:
      self._UpdateLocation(True, output)
      self.PrintMarkovChainTransition(element, output)

  def _UpdateLocation(self, inTransition, output):
    if self._inTransition and not inTransition:
      self._inTransition = False
      self._depth -= 1
    elif not self._inTransition and inTransition:
      self._inTransition = True
      self.PrintIndentation(output)
      self.PrintKeyword(AndreyType.KEYWORD_TRANSITION, output)
      output.write("\n")
      self._depth += 1

# 3.2) Blocks and classes
# -----------------------

  def _PrintBlockBody(self, block, output):
    self._inTransition = False
    self._depth += 1
    for element in block.GetElements():
      self.PrintElement(element, output)
    self._UpdateLocation(False, output)
    self._depth -= 1

# 3.3) Specific elements
# ----------------------

  def PrintMarkovChainState(self, state, output):
    self.PrintIndentation(output)
    self.PrintKeyword(AndreyType.KEYWORD_STATE, output)
    output.write(" ")
    self.PrintPath(state.GetPath(), output)
    self.PrintAttributes(state, output)
    output.write(";\n")

  def PrintMarkovChainEvent(self, event, output):
    self.PrintIndentation(output)
    self.PrintKeyword(AndreyType.KEYWORD_EVENT, output)
    output.write(" ")
    self.PrintPath(event.GetPath(), output)
    self.PrintAttributes(event, output)
    output.write(";\n")

  def PrintMarkovChainParameter(self, parameter, output):
    self.PrintIndentation(output)
    self.PrintKeyword(AndreyType.KEYWORD_PARAMETER, output)
    output.write(" ")
    self.PrintPath(parameter.GetPath(), output)
    output.write(" = ")
    self.PrintStochasticExpression(parameter.GetDefinition(), output)
    output.write(";\n")

  def PrintMarkovChainReward(self, reward, output):
    self.PrintIndentation(output)
    self.PrintKeyword(AndreyType.KEYWORD_REWARD, output)
    output.write(" ")
    self.PrintPath(reward.GetPath(), output)
    output.write(" = ")
    self.PrintStochasticExpression(reward.GetDefinition(), output)
    output.write(";\n")

  def PrintMarkovChainTransition(self, transition, output):
    self.PrintIndentation(output)
    self.PrintPath(transition.GetEventPath(), output)
    output.write(": ")
    self.PrintPath(transition.GetSourceStatePath(), output)
    output.write(" -> ")
    self.PrintPath(transition.GetTargetStatePath(), output)
    output.write(";\n")

