# S2MLCore version 1.0.0
# S2MLLexer.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Token
# 3) Lexer

# 1) Imported modules
# -------------------

import string
import sys
import re

from S2MLType import *

# 2) Token
# --------

class S2MLToken:
  def __init__(self, type, string, fileName, lineNumber, columnNumber):
    self.type = type
    self.string = string
    self.fileName = fileName
    self.lineNumber = lineNumber
    self.columnNumber = columnNumber

  @staticmethod
  def GetNameOfType(type):
    return type

  def GetType(self):
    return self.type
  def GetString(self):
    return self.string
  def GetFileName(self):
    return self.fileName
  def GetLineNumber(self):
    return self.lineNumber
  def GetColumnNumber(self):
    return self.columnNumber

# 3) Lexer
# --------

# 3.1) Creator
# ------------

class S2MLLexer:
  STATE_END_OF_FILE = 0
  STATE_INITIAL = 1
  STATE_IDENTIFIER = 10
  STATE_PATH_1 = 11
  STATE_PATH_2 = 12
  STATE_NUMBER_1 = 20
  STATE_NUMBER_2 = 21
  STATE_NUMBER_3 = 22
  STATE_NUMBER_4 = 23
  STATE_NUMBER_5 = 24
  STATE_NUMBER_6 = 25
  STATE_NUMBER_7 = 26
  STATE_NUMBER_8 = 27
  STATE_SINGLE_LINE_COMMENT = 40
  STATE_MULTILINE_COMMENT_1 = 50
  STATE_MULTILINE_COMMENT_2 = 51

  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._actions = dict()
    self.RegisterActions()
    self._codesOfKeywords = dict()
    self._stringsOfCodes = dict()
    self.RegisterTokens()
    self.RegisterKeywords()
    self.keepComments = False
    self.fileName = None
    self.file = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.line = None
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.tokens = []
    self.state = -1

# 3.2) Management of actions
# --------------------------

  def RegisterActions(self):
    self.RegisterAction("(", S2MLLexer.ActionLPAR)
    self.RegisterAction(")", S2MLLexer.ActionRPAR)
    self.RegisterAction("{", S2MLLexer.ActionLBRACE)
    self.RegisterAction("}", S2MLLexer.ActionRBRACE)
    self.RegisterAction(";", S2MLLexer.ActionSemicolon)
    self.RegisterAction(":", S2MLLexer.ActionColon)
    self.RegisterAction(",", S2MLLexer.ActionComma)
    self.RegisterAction("=", S2MLLexer.ActionEqual)
    self.RegisterAction("+", S2MLLexer.ActionPlus)
    self.RegisterAction("-", S2MLLexer.ActionMinus)
    self.RegisterAction("/", S2MLLexer.ActionSlash)

  def RegisterAction(self, c, action):
    self._actions[c] = action

  def GetAction(self, c):
    return self._actions.get(c, None)

  def ActionLPAR(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_LPAR)

  def ActionRPAR(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_RPAR)

  def ActionLBRACE(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_LBRACE)

  def ActionRBRACE(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_RBRACE)

  def ActionSemicolon(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_SEMICOLON)

  def ActionColon(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_COLON)

  def ActionComma(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_COMMA)

  def ActionEqual(self, c):
    self.tokenString = c
    self.NewToken(S2MLType.TOKEN_EQUAL)

  def ActionPlus(self, c):
    c = self.GetChar()
    if c.isdigit():
      self.tokenString = "+" + c
      self.state = S2MLLexer.STATE_NUMBER_1
    else:
      self.tokenString = "+"
      self.PutCharBack(c)
      self.NewToken(S2MLType.TOKEN_UNKNOWN)

  def ActionMinus(self, c):
    c = self.GetChar()
    if c.isdigit():
      self.tokenString = "-" + c
      self.state = S2MLLexer.STATE_NUMBER_1
    else:
      self.tokenString = "-"
      self.PutCharBack(c)
      self.NewToken(S2MLType.TOKEN_UNKNOWN)

  def ActionSlash(self, c):
    c = self.GetChar()
    if c=="/":
      self.tokenString = "//"
      self.state = S2MLLexer.STATE_SINGLE_LINE_COMMENT
    elif c=="*":
      self.tokenString = "/*"
      self.state = S2MLLexer.STATE_MULTILINE_COMMENT_1
    else:
      self.tokenString = "/"
      self.PutCharBack(c)
      self.NewToken(S2MLType.TOKEN_UNKNOWN)

# 3.3) Management of keywords
# ---------------------------

  def RegisterTokens(self):
    self.RegisterToken("(", S2MLType.TOKEN_LPAR)
    self.RegisterToken(")", S2MLType.TOKEN_RPAR)
    self.RegisterToken("{", S2MLType.TOKEN_LBRACE)
    self.RegisterToken("}", S2MLType.TOKEN_RBRACE)
    self.RegisterToken(";", S2MLType.TOKEN_SEMICOLON)
    self.RegisterToken(":", S2MLType.TOKEN_COLON)
    self.RegisterToken(",", S2MLType.TOKEN_COMMA)
    self.RegisterToken("=", S2MLType.TOKEN_EQUAL)
    self.RegisterToken("number", S2MLType.TOKEN_NUMBER)
    self.RegisterToken("identifier", S2MLType.TOKEN_IDENTIFIER)
    self.RegisterToken("path", S2MLType.TOKEN_PATH)
    self.RegisterToken("comment", S2MLType.TOKEN_COMMENT)

  def RegisterToken(self, string, code):
    self._stringsOfCodes[code] = string

  def RegisterKeywords(self):
    # S2ML
    self.RegisterKeyword("domain", S2MLType.KEYWORD_DOMAIN)
    self.RegisterKeyword("block", S2MLType.KEYWORD_BLOCK)
    self.RegisterKeyword("class", S2MLType.KEYWORD_CLASS)
    self.RegisterKeyword("end", S2MLType.KEYWORD_END)
    self.RegisterKeyword("clones", S2MLType.KEYWORD_CLONES)
    self.RegisterKeyword("as", S2MLType.KEYWORD_AS)
    self.RegisterKeyword("embeds", S2MLType.KEYWORD_EMBEDS)
    self.RegisterKeyword("extends", S2MLType.KEYWORD_EXTENDS)
    self.RegisterKeyword("parameter", S2MLType.KEYWORD_PARAMETER)
    # Expressions
    self.RegisterKeyword("add", S2MLType.KEYWORD_ADD)
    self.RegisterKeyword("sub", S2MLType.KEYWORD_SUB)
    self.RegisterKeyword("mul", S2MLType.KEYWORD_MUL)
    self.RegisterKeyword("div", S2MLType.KEYWORD_DIV)
    self.RegisterKeyword("neg", S2MLType.KEYWORD_NEG)
    self.RegisterKeyword("min", S2MLType.KEYWORD_MIN)
    self.RegisterKeyword("max", S2MLType.KEYWORD_MAX)
    self.RegisterKeyword("false", S2MLType.KEYWORD_FALSE)
    self.RegisterKeyword("true", S2MLType.KEYWORD_TRUE)
    self.RegisterKeyword("and", S2MLType.KEYWORD_AND)
    self.RegisterKeyword("or",  S2MLType.KEYWORD_OR)
    self.RegisterKeyword("not", S2MLType.KEYWORD_NOT)
    self.RegisterKeyword("eq",  S2MLType.KEYWORD_EQ)
    self.RegisterKeyword("df",  S2MLType.KEYWORD_DF)
    self.RegisterKeyword("lt",  S2MLType.KEYWORD_LT)
    self.RegisterKeyword("gt",  S2MLType.KEYWORD_GT)
    self.RegisterKeyword("leq", S2MLType.KEYWORD_LEQ)
    self.RegisterKeyword("count", S2MLType.KEYWORD_COUNT)
    self.RegisterKeyword("if", S2MLType.KEYWORD_IF)

  def RegisterKeyword(self, keyword, code):
    self._codesOfKeywords[keyword] = code
    self._stringsOfCodes[code] = keyword

  def GetIdentifierType(self, identifier):
    return self._codesOfKeywords.get(identifier, S2MLType.TOKEN_IDENTIFIER)

  def GetCodeOfKeyword(self, keyword):
    return self._codesOfKeywords.get(keyword, None)

  def GetStringOfCode(self, code):
    return self._stringsOfCodes.get(code, None)

# 3.4) Main loop
# --------------

  def Initialize(self, fileName, file):
    self.fileName = fileName
    self.file = file
    self.tokens = []
    self.line = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.GetNewLine()
    self.state = S2MLLexer.STATE_INITIAL
    return 0

  def Finalize(self):
    pass
     # self.file.close()

  def GetToken(self, tokenIndex):
    if tokenIndex<len(self.tokens):
      return self.tokens[tokenIndex]
    while self.state!=S2MLLexer.STATE_END_OF_FILE and tokenIndex>=len(self.tokens):
      if self.state==S2MLLexer.STATE_INITIAL:
        c = self.GetChar()
        self.tokenLineNumber = self.lineNumber
        self.tokenColumnNumber = self.columnNumber-1
        if c==None:
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isspace():
          continue
        elif c.isalpha() or c=="_":
          self.tokenString = c
          self.state = S2MLLexer.STATE_IDENTIFIER
        elif c.isdigit():
          self.tokenString = c
          self.state = S2MLLexer.STATE_NUMBER_2
        else:
          action = self.GetAction(c)
          if action==None:
            self.tokenString = c
            self.NewToken(S2MLType.TOKEN_UNKNOWN)
          else:
            action(self, c)

      # -----------------------------------------------------
      # Identifiers & Paths
      # -----------------------------------------------------
      elif self.state==S2MLLexer.STATE_IDENTIFIER:
        c = self.GetChar()
        if c.isdigit() or c.isalpha() or c=="_" or c=="-":
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = S2MLLexer.STATE_PATH_1
        else:
          self.PutCharBack(c)
          type = self.GetIdentifierType(self.tokenString)
          self.NewToken(type)
          self.state = S2MLLexer.STATE_INITIAL

      elif self.state==S2MLLexer.STATE_PATH_1:
        c = self.GetChar()
        if c.isalpha() or c=="_" or c=="-":
          self.tokenString += c
          self.state = S2MLLexer.STATE_PATH_2
        else:
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_PATH_2:
        c = self.GetChar()
        if c.isdigit() or c.isalpha() or c=="_" or c=="-":
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = S2MLLexer.STATE_PATH_1
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_PATH)
          self.state = S2MLLexer.STATE_INITIAL

      # -----------------------------------------------------
      # Numbers
      # -----------------------------------------------------
      elif self.state==S2MLLexer.STATE_NUMBER_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_2
        elif c==".":
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_8
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_NUMBER_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_NUMBER)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_3
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_NUMBER)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_NUMBER_3:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_4
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_NUMBER_4:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_NUMBER)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_NUMBER)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_NUMBER_5:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_7
        elif c=="+" or c=="-":
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_6
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_NUMBER_6:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_7
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_NUMBER_7:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_NUMBER)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_NUMBER)
          self.state = S2MLLexer.STATE_INITIAL
      elif self.state==S2MLLexer.STATE_NUMBER_8:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLLexer.STATE_NUMBER_4
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLType.TOKEN_UNKNOWN)
          self.state = S2MLLexer.STATE_INITIAL

      # -----------------------------------------------------
      # Single line comments
      # -----------------------------------------------------
      elif self.state==S2MLLexer.STATE_SINGLE_LINE_COMMENT:
        c = self.GetChar()
        if c==None:
          if self.keepComments:
            self.NewToken(S2MLType.TOKEN_COMMENT)
          self.state = 0
        elif c=="\n":
          if self.keepComments:
            self.NewToken(S2MLType.TOKEN_COMMENT)
          self.state = 1
        else:
          self.tokenString += c

      # -----------------------------------------------------
      # Multiline comments
      # -----------------------------------------------------
      elif self.state==S2MLLexer.STATE_MULTILINE_COMMENT_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_NON_TERMINATED_COMMENT)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c=="*":
          self.tokenString += c
          self.state = S2MLLexer.STATE_MULTILINE_COMMENT_2
        else:
          self.tokenString += c
      elif self.state==S2MLLexer.STATE_MULTILINE_COMMENT_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLType.TOKEN_NON_TERMINATED_COMMENT)
          self.state = S2MLLexer.STATE_END_OF_FILE
        elif c=="*":
          self.tokenString += c
          c = self.GetChar()
        elif c=="/":
          self.tokenString += c
          if self.keepComments:
            self.NewToken(S2MLType.TOKEN_COMMENT)
          self.state = S2MLLexer.STATE_INITIAL
        else:
          self.tokenString += c
          self.state = S2MLLexer.STATE_MULTILINE_COMMENT_1

    if tokenIndex>=len(self.tokens):
      return None
    return self.tokens[tokenIndex]

# 3.5) Auxiliary functions
# ------------------------

  def NewToken(self, type):
    token = S2MLToken(type, self.tokenString, self.fileName, self.tokenLineNumber, self.tokenColumnNumber)
    self.tokens.append(token)

  def GetChar(self):
    if len(self.readChars)!=0:
      return self.readChars.pop()
    if self.columnNumber==len(self.line):
      self.GetNewLine()
      if self.endOfFile:
        return None
      return self.GetChar()
    c = self.line[self.columnNumber]
    self.columnNumber += 1
    return c

  def PutCharBack(self, c):
    self.readChars.append(c)

  def GetNewLine(self):
    self.line = self.file.readline()
    if self.line=="":
      self.endOfFile = 1
      return
    self.lineNumber += 1
    self.columnNumber = 0
    self.line = self.line.rstrip()
    self.line += "\n"

  def ScanFile(self, fileName):
    error = self.Initialize(fileName)
    if error!=0:
      return
    tokenIndex = 0
    while True:
      token = self.GetToken(tokenIndex)
      if token==None:
        break
      tokenIndex += 1
    self.Finalize()

  def PrintTokens(self, outputFile):
    for token in self.tokens:
      outputFile.write(str(token.lineNumber) + "\t")
      outputFile.write(str(token.columnNumber) + "\t")
      outputFile.write(S2MLToken.GetNameOfType(token.type) + "\t")
      outputFile.write(token.string + "\n")
